import os
import xml.etree.ElementTree as ET
import hashlib

from ..constants import OSM_CACHE_DIR, OSM_TYPES


def get_cache_key(query: str) -> str:
    """
    Generate a unique cache key for the query using a hash function.
    This ensures that queries with different parameters are cached separately.
    """
    return hashlib.md5(query.encode('utf-8')).hexdigest()

class CachingStrategyBase:
    """
    Base class for implementing caching strategies.
    """
    def get(self, key):
        """Retrieve the cached data associated with the provided key."""
        raise NotImplementedError('Subclass should implement get')

    def set(self, key, value):
        """Store data in the cache with the specified key."""
        raise NotImplementedError('Subclass should implement set')

    def set_hollow(self, key, cell: tuple, osm_types: OSM_TYPES,
                    selector: str, conditions=[], out='center'):
        """Create a hollow (empty) cache entry with a specific key."""
        raise NotImplementedError('Subclass should implement set_hollow')

    def fill_hollow(self, key, value):
        """Fill in the cache for an existing hollow entry."""
        raise NotImplementedError('Subclass should implement fill_hollow')

    def close(self):
        """Clean up or close any resources used by the caching strategy."""


class XMLCache(CachingStrategyBase):
    """
    A caching strategy that stores and retrieves data in XML format.
    """
    def __init__(self, cache_dir=OSM_CACHE_DIR):
        # Add the class name as a suffix to the directory
        self._cache_dir = f'{cache_dir}'
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)

    def _filename(self, key):
        return os.path.join(self._cache_dir, f'{key}.xml')

    def get(self, key):
        """Retrieve XML data from the cache and parse it as an ElementTree."""
        filename = self._filename(key)
        if os.path.exists(filename):
            try:
                # Parse and return the cached XML data
                tree = ET.parse(filename)
                return tree.getroot()  # Return the root element of the parsed XML
            except ET.ParseError:
                return None
        return None

    def set(self, key, value):
        """Save the XML data as an ElementTree to the cache."""
        filename = self._filename(key)
        tree = ET.ElementTree(value)  # value is expected to be an ElementTree root element
        try:
            with open(filename, 'wb') as file:
                tree.write(file, encoding='utf-8', xml_declaration=True)
        except IOError as e:
            raise IOError(f"Error writing to cache file: {filename} - {e}") from e

    def set_hollow(self, key, cell: tuple, osm_types: OSM_TYPES,
                    selector: str, conditions=[], out='center'):
        """Create an empty placeholder cache entry for a future fill."""
        hollow_key = f'hollow_{key}'
        filename = self._filename(hollow_key)
    
        # Create the root element <cache>
        root = ET.Element("params")
        # Add sub-elements with provided values
        ET.SubElement(root, "key").text = key
        ET.SubElement(root, "cell").text = f"({cell[0]}, {cell[1]})"
        ET.SubElement(root, "osm_types").text = ','.join(osm_types)
        ET.SubElement(root, "selector").text = selector
        ET.SubElement(root, "conditions").text = ','.join(conditions) if conditions else "none"
        ET.SubElement(root, "out").text = out

        # Create an ElementTree object from the root
        tree = ET.ElementTree(root)

        # Write the XML to the file
        with open(filename, 'wb') as file:
            tree.write(file, encoding='utf-8', xml_declaration=True)

    def close(self):
        """Cleanup method, if needed."""
        pass

class CachingStrategy:
    """
    A class to manage different caching strategies.
    """
    __strategy = XMLCache()  # Default caching strategy
    __strategies = {
        'XML': XMLCache,
    }

    @classmethod
    def use(cls, strategy_name='XML', **kwargs):
        if cls.__strategy:
            cls.__strategy.close()

        strategy_class = cls.__strategies.get(strategy_name)
        if not strategy_class:
            raise ValueError(f"Unknown caching strategy: {strategy_name}")

        cls.__strategy = strategy_class(**kwargs)
        return cls.__strategy

    @classmethod
    def get(cls, key):
        return cls.__strategy.get(key)

    @classmethod
    def set(cls, key, value):
        cls.__strategy.set(key, value)

    @classmethod
    def set_hollow(cls, key, cell: tuple, osm_types: OSM_TYPES,
                    selector: str, conditions=[], out='center'):
        """Create a hollow cache entry."""
        cls.__strategy.set_hollow(key, cell, osm_types, selector, conditions, out)

    @classmethod
    def fill_hollow(cls, key, value):
        """Fill in the hollow cache entry with actual data."""
        cls.__strategy.fill_hollow(key, value)
